// RC4 sequential KR with concatenated IVs

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include "SFMT.h" // PRNG

#define SWAPI(X,Y) \
{                  \
  int __tmp;       \
  __tmp = (X);     \
  (X)   = (Y);     \
  (Y)   = __tmp;   \
}

/*
 * Initialises the RC4 permutation (state) from the given key,
 * up to step upto.
 * Returns the value of the "j" index at the end of the initialisation.
 * A full init. (256 steps) is performed if upto is set to -1.
 * 
 * /!\
 * WARNING ABOUT NOTATIONS: The permutation S_n is obtained
 * after a call with upto set to n + 1.
 * This is because the initial state is S_{-1}.
 */
int rc4_sinit(uint8_t state[256], uint8_t key[], int keylen, int upto)
{
  int i, j = 0;
  upto = upto == - 1 ? 256 : upto;

  for (i = 0; i < 256; i++)
    state[i] = i;

  for (i = 0; i < upto; i++)
  {
    j += state[i] + key[i % keylen];
    j &= 0xff;
    SWAPI(state[i], state[j]);
  }

  return j; // useful in the KR attack
}

/*
 * Returns the first byte of the keystream obtained
 * from the state
 */
uint8_t rc4_get_z0(uint8_t state[256])
{
  return state[(state[1] + state[state[1]]) & 0xff];
}

/*
 * Sets the IV of key (the 3 first words) so that
 * it can be used to guess the secret key word
 * target, counting from zero (i.e. key[target + 3]).
 * (Simplest case --- Wagner scenario)
 */
void rz_prepare_3w_iv(uint8_t key[], int keylen, int target, sfmt_t *sfmt)
{
  // IV is [target + 3 | 255 | nonce]
  key[0] = target + 3;
  key[1] = 255;
  key[2] = sfmt_genrand_uint32(sfmt) & 0xff; // too bad for 32/8

  return;
}

/*
 * Performs the initialisation with the key up to S_{target + 2}, then
 * checks if the state is good for use in statistical KR
 * (i.e. s[0] = target + 3, s[1] = 0).
 */
int iv_okay(uint8_t key[], int keylen, int target)
{
  uint8_t state[256];
  int i, j = 0;

  assert(target + 3 <= keylen);

  // init phase 1
  for (i = 0; i < 256; i++)
    state[i] = i;

  // init phase 2
  for (i = 0; i < target + 3; i++)
  {
    j += state[i] + key[i];
    j &= 0xff;
    SWAPI(state[i], state[j]);
  }

  if ((state[0] == (target + 3)) && (state[1] == 0))
    return 1;
  return 0;
}

/*
 * Randomly initializes key (except the 3 first words
 * which are really IVs) of size keylen bytes
 */
void rz_keygen(uint8_t key[], int keylen, sfmt_t *sfmt)
{
  int i;

  for (i = 3; i < keylen; i++)
    key[i] = sfmt_genrand_uint32(sfmt); // I know, subopt.
}

/*
 * Key recovery attack in Wagner's scenario, §7.1
 */
void get_back_the_key(int keylen, sfmt_t *sfmt)
{
  // TODO
}

int main(int argc, char **argv)
{
  uint64_t seed;
  sfmt_t sfmt;

  if (argc != 2)
  {
    printf("usage: rc4KR <seed>\n");
    return 1;
  }

  seed = atoi(argv[1]);
  sfmt_init_gen_rand(&sfmt, seed);

  get_back_the_key(8, &sfmt);

  return 0;
}
